/**********************************************************************
 ***                                                                ***
 ***                                                                ***
 ***        HDLOOKUP - Hard disk serialization, version 1.5         ***
 ***                                                                ***
 ***                                                                ***
 ***    This program is used to check a unique serial               ***
 ***    number, using 1st hard disk characteristics                 ***
 ***    such as physical parameters and partition table.            ***
 ***    Data are encrypted and stored in a special file,            ***
 ***    which is created earlier with HDLOCK and is used to         ***
 ***    compare these data with the data re-constructed by          ***
 ***    the program every time it is executed.                      ***
 ***                                                                ***
 ***    Exit codes are:                                             ***
 ***        0 = Success (data identical)                            ***
 ***        1 = Int-13h (func=15h - identify disk) failure          ***
 ***        2 = Disk 80h does not exist (no 1st hard disk)          ***
 ***        3 = Disk 80h not a hard disk                            ***
 ***        4 = Int-13h (func=02h - read disk sector) failure       ***
 ***        5 = Memory error                                        ***
 ***        6 = Error creating data files                           ***
 ***        7 = Error reading data files                            ***
 ***        8 = Data not identical (validation failure)             ***
 ***       10 = No action (/? switch or invalid options)            ***
 ***                                                                ***
 ***                                                                ***
 ***                            Harris Georgiou, 1996.              ***
 ***                                                                ***
 **********************************************************************/



#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <mem.h>
#include <string.h>

#define         VERNUM_MAJ      (char)1
#define         VERNUM_MIN      (char)5

#define         DRVNUM          0x80        /* 80h = 1st hard disk */
#define         DRVINTNUM       0x41        /* 1st hard disk description table */

#define         MAXSTRLEN       128         /* maximum input string length */

#define         HDDTSIZE        72          /* =(4+4*16+2)+2 , multiplicant of 8 */
#define         SECTSIZE        512         /* sector size (bytes) */

#define         EOFCHAR         0x1a        /* ^Z = end-of-file character */

#define         DEFPASSWD       106         /* default password code (byte) */
#define         ENCRPARAM       53          /* shift code parameter */
#define         MAXBYTEVAL      256         /* maximum byte value */

#define         DEFTXTSIGN      "SIGN.TXT"  /* default signature text file (never used) 
                                                    may be used in future versions */
#define         DEFBINSIGN      "SIGN.BIN"  /* default binary signature file */


int     quiet_flag=0;           /* global variable, used for "quiet" execution (no echo) */


/*** main data structure for HDDT record ***/
struct  Thddt_data  {    
                unsigned int    vernum_major;
                unsigned int    vernum_minor;
                
                unsigned char   date_day;
                unsigned char   date_month;
                unsigned int    date_year;

                unsigned long   passwd;

                unsigned char   data[HDDTSIZE]; };



/*** Read 1st hard disk parameters (use BIOS) ***/
int get_hd_param( unsigned char *hddt )
{
    union REGS            inregs, outregs;
    unsigned char   heads=0, sec_trk=0;
    unsigned int    cyls=0, vectseg, vectoff;

    
    inregs.h.ah=0x15;           /* get disk type */
    inregs.h.dl=0x80;
    int86( 0x13, &inregs, &outregs );
    
    if (outregs.x.cflag != 0)       /* general int-13h error */
        return(1);
    else if (outregs.h.ah == 0)     /* drive does not exist */
        return(2);
    else if (outregs.h.ah != 3)     /* not a hard disk */
        return(3);
    else  /* outregs.h.ah == 3 */   /* OK */
    {
        vectseg = FP_SEG(getvect(0x41));       /* get 1st hard disk information */
        vectoff = FP_OFF(getvect(0x41));       /* directly from its description table */
        heads = peekb( vectseg, vectoff+0x02 );
        cyls = peek( vectseg, vectoff+0x00 );
        sec_trk = peekb( vectseg, vectoff+0x0e );

        memcpy( &hddt[0], &heads, 1 );
        memcpy( &hddt[1], &cyls, 2 );
        memcpy( &hddt[3], &sec_trk, 1 );
    
        return(0);
    }
}


/*** Read 1st hard disk partition table ***/
int get_hd_prtable( unsigned char *hddt )
{
    union REGS          inregs, outregs;
    struct SREGS        s_ioregs;
    unsigned char       buff[SECTSIZE];


    memset( hddt, 0, HDDTSIZE );
    
/*****   Warning:   The following (HD reset) routine may 
                    hang-up the system on some old AT machines   *****
    inregs.h.ah=0x0d;                   * reset hard disk (00h for floppies) *
    inregs.h.dl=0x80;                   * use 1st hard disk *
    int86( 0x13, &inregs, &outregs );
*****/    

    inregs.h.ah=0x02;                   /* read sector from disk */
    inregs.h.al=0x01;                   /* read only 1 sector */
    inregs.h.dl=0x80;                   /* specify (hard) disk number */
    inregs.h.ch=0x00;                   /* cylinder 0 (low bits) */
    inregs.h.cl=0x01;                   /* sector 1 (cyl 0, high bits) */
    inregs.h.dh=0x00;                   /* head 0 (1st side) */
    s_ioregs.es=FP_SEG(buff);           /* read buffer segment */
    inregs.x.bx=FP_OFF(buff);           /* read buffer offset */
    int86x( 0x13, &inregs, &outregs, &s_ioregs );

    if (outregs.x.cflag != 0) 
        return(1);
    else
    {
        memcpy( &hddt[4], &buff[0x1be], 4*16+2 );
        if (memcmp(&hddt[4],&buff[0x1be],4*16+2)!=0)
            return(2);
        else
            return(0);
    }
}


/*** Add an EOF character at the end of a text file ***/
void add_eof_char( FILE *infile, FILE *outfile )
{
    unsigned char   buff=0;
    
    
    while (!feof(infile))
    {
        fread( &buff, sizeof(char), 1, infile );
        fwrite( &buff, sizeof(char), 1, outfile );
    }
    buff = EOFCHAR;
    fwrite( &buff, sizeof(char), 1, outfile );
}


/*** General file copying subroutine ***/
void copyfile( FILE *infile, FILE *outfile )
{
    unsigned char   buff=0;
    
    
    while (!feof(infile))
    {
        fread( &buff, sizeof(char), 1, infile );
        fwrite( &buff, sizeof(char), 1, outfile );
    }
}


/*** Calculate a file's length in bytes (similar to "filelength") ***/
unsigned long countfilelen( FILE *infile )
{
    unsigned char   buff=0;
    unsigned long   len=0L;
    fpos_t          pos;

    fgetpos( infile, &pos );
    fseek( infile, 0L, SEEK_SET );
    
    len=0L;
    while (!feof(infile))  
    {
        fread( &buff, sizeof(char), 1, infile );
        len++;
    }
    
    fsetpos( infile, &pos );
    return( len );
}


/*** Decrypt a single byte (uses XOR and value shift) ***/
unsigned char decrypt_byte( unsigned char origbyte, unsigned char pwd )
{
    int         decrbyte=0;

    decrbyte = origbyte + ENCRPARAM;
    if ( decrbyte>(MAXBYTEVAL-1) )
        decrbyte = (decrbyte - MAXBYTEVAL) ^ pwd;
    else
        decrbyte = decrbyte ^ pwd;

    return( (unsigned char)decrbyte );
}


/*** Decrypt memory block with password ***/
void unscramble_memblock( unsigned char *buff, unsigned int length,
                            unsigned long passwd )
{
    unsigned int        count=0;


    for ( count=0; count<length; count++ )
    {
        buff[count] = decrypt_byte( buff[count], (char)passwd );      /* --> decryption function */
    }
}


/*** Basic serialization file read ***/
int read_sign_file( char *signfilename, unsigned char *hddt )
{
    FILE            *signfile;
    struct Thddt_data      hddtrec;
    fpos_t          fpos;
    unsigned long   pos=0;


    if (!quiet_flag) printf("Reading serialization file... (%s)\n",signfilename);

/*** Initialization checks ***/
    if ( (signfile = fopen(signfilename,"rb"))==NULL )
    {
        fclose( signfile );
        return(1);
    }

/*** read HDDT record from file to memory ***/
    fread( &pos, sizeof(long), 1, signfile );
    fpos = (fpos_t)pos;
    fsetpos( signfile, &fpos );
    fread( &hddtrec, sizeof(struct Thddt_data), 1, signfile );

/*** show HDDT record information block ***/
    if (!quiet_flag) printf("Version: %d.%d\t\t",hddtrec.vernum_major,hddtrec.vernum_minor);
    if (!quiet_flag) printf("Created: %d/%d/%d",hddtrec.date_day,hddtrec.date_month,hddtrec.date_year);

    unscramble_memblock( hddtrec.data, HDDTSIZE, hddtrec.passwd );
    memcpy( hddt, hddtrec.data, HDDTSIZE );
    if (!quiet_flag) printf("\nSerialization data decrypted.\n");

    if (!quiet_flag) printf("Serialization file read successfully!\n");
    fclose( signfile );
    return(0);
}


/*** Compare serialization (file) data with re-calculated ***/
int verify_hddt( unsigned char *read_hddt )
{
    unsigned char   retrv_hddt[HDDTSIZE];

    get_hd_param( retrv_hddt );
    get_hd_prtable( retrv_hddt );
    return( memcmp( read_hddt, retrv_hddt, HDDTSIZE ) );
}


/*** Display program usage ***/
void usage( void )
{
    if (!quiet_flag) printf("    Usage:   HDLOOKUP   [ <sign_bin_file>  [ /Q ]  ]\n\n");
    if (!quiet_flag) printf("\t\t<sign_bin_file> = serialization data file\n");
    if (!quiet_flag) printf("\t\t                  (if left null, default is used)\n");
    if (!quiet_flag) printf("\t\t/Q                use quiet mode (no echo)\n\n");
    if (!quiet_flag) printf("    Example:   hdlookup  serial.bin  /Q\n\n");
}



/*** MAIN PROGRAM ***/
int main( int argc, char **argv )
{
    unsigned char   heads=0, sec_trk=0;
    unsigned char   hddt[HDDTSIZE];
    unsigned int    cyls=0;
    int             res=0;
    char            resp[MAXSTRLEN];
    char            *defsignfilename=DEFBINSIGN;
    char            signfilename[MAXSTRLEN];

        
    if ( (argc==3)&&((cyls=stricmp(argv[2],"/Q"))==0) )
    {
        quiet_flag = 1;
        argc--;
    }
    else
        quiet_flag = 0;

    if (!quiet_flag) printf("Hard disk recognizer - version %d.%d, Harris Georgiou, Copyright (c) 1996.\n\n",
                                                        VERNUM_MAJ,VERNUM_MIN);
    
    if (argc>=2)
        if ( (stricmp(argv[1],"/?")==NULL)||(stricmp(argv[1],"/h")==NULL) )
        {
            usage();
            return(10);
        }

    switch( argc )
    {
        case 1: strncpy( signfilename, defsignfilename, MAXSTRLEN );
                break;
        case 2: strncpy( signfilename, argv[1], MAXSTRLEN );
                break;
       default: usage();
                return(10);
    }
    
    if (!quiet_flag) printf("Identifying system disks...\n");
    res = get_hd_param( hddt );
    switch( res )
    {
        case 1: if (!quiet_flag) printf("ERROR: Couldn't call disk routine identification!\n");
                return(1);
        case 2: if (!quiet_flag) printf("ERROR: Drive %Xh does not exist!\n",DRVNUM);
                return(2);
        case 3: if (!quiet_flag) printf("ERROR: Drive %Xh is not a hard disk!\n",DRVNUM);
                return(3);
        case 0:
       default: if (!quiet_flag) printf("Drive %Xh detected as hard disk:\n",DRVNUM);
                heads=hddt[0];
                memcpy( &cyls, &hddt[1], sizeof(int) );
                sec_trk=hddt[3];
                if (!quiet_flag) printf("\t#heads=%u   #cyls=%u   #sec/cyl=%u\n",heads,cyls,sec_trk);
                break;
    }
    
    if (!quiet_flag) printf("\n");
    
    if (!quiet_flag) printf("Reading system disk information...");
    res=get_hd_prtable( hddt );
    switch( res ) 
    {
        case 1: if (!quiet_flag) printf("\nERROR: Couldn't call disk routine for reading!\n");
                return(4);
        case 2: if (!quiet_flag) printf("\nERROR: Couldn't copy memory blocks correctly!\n");
                return(5);
        case 0:
       default: if (!quiet_flag) printf(" OK\n");  break;
    }  /*** end switch ***/

/*** Serialization file read & verification ***/
    if (!quiet_flag) printf("\nTrying to recover data...\n\n");

    if ( read_sign_file(signfilename,hddt) )
    {
        if (!quiet_flag) printf("ERROR: Cannot open data file\n");
        return(7);
    }
    else        
    {
        if ( verify_hddt(hddt) != 0 )
        {
            if (!quiet_flag) printf("\nWARNING: Data NOT verified correctly!\n");
            return(8);
        }
        else
        {
            if (!quiet_flag) printf("\nData verified correctly!\n");
            return(0);
        }
    }

}

